/* This file is part of Netsukuku
 * (c) Copyright 2004 Andrea Lo Pumo aka AlpT <alpt@freaknet.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#define MAX_RADAR_SCANS		10
#define MAX_RADAR_WAIT		5	/*How much we wait to store the received ECHO_REPLY pkts
					  and then to close the current radar session*/
#define RTT_DELTA		1000	/*If the change delta of the new rtt is >= RTT_DELTA, 
					  the qspn_q.send_qspn will be set. (It's in millisec)*/

#ifdef DEBUG		
#undef MAX_RADAR_WAIT
#define MAX_RADAR_WAIT          3
#endif

int max_radar_wait;
int radar_scans;			/*How many ECHO_ME pkts we sent*/
int radar_scan_mutex;			/*A flag to see if we are already doing a scan*/
int my_echo_id;			
u_char send_qspn_now[MAX_LEVELS];	/*Shall we send the qspn in level? If yes 
					  send_qspn_now[level] is != 0*/
int hook_retry;				/*If we've seen, while hooking, a node who was trying 
					  to hook before us, `hook_retry' is set to 1.*/

int total_radars;			/*Stupid statistic*/

#define RADQ_VOID_RNODE		0
#define RADQ_EXT_RNODE		1

struct radar_queue
{
	struct radar_queue *next;
	struct radar_queue *prev;

	inet_prefix	ip;			/*Node's ip*/
	map_node       *node;			/*The node we are pinging*/
	quadro_group	quadg;			/*Node's data for the ext_map*/
	u_short		flags;
				
	char 		pings;			/*The total ECHO_ME pkts received from this node*/
	char 		pongs;			/*The total pongs (ECHO_REPLY) received from this node*/
	struct timeval 	rtt[MAX_RADAR_SCANS];	/*The round rtt of each pong*/
	struct timeval 	final_rtt;		/*When all the rtt is filled, or when MAX_RADAR_WAIT is expired,
						  final_rtt will keep the average of all the rtts */
};
struct radar_queue *radar_q;	/*the start of the linked list of radar_queue*/
int radar_q_counter;

struct timeval scan_start;	/*the start of the scan*/

/* 
 * The ECHO_ME pkt:
 * It is just a normal pkt which contains in the body (pkt.msg) one 
 * u_char echo_scans_count, var. This is the number of scans sent.
 */

void init_radar(void);
void close_radar(void);
void reset_radar(void);
void free_new_node(void);

struct radar_queue *find_node_radar_q(map_node *node);
struct radar_queue *find_ip_radar_q(inet_prefix *ip);
int count_hooking_nodes(void);

void final_radar_queue(void);
void radar_update_map(void);

struct radar_queue *add_radar_q(PACKET pkt);
int radar_exec_reply(PACKET pkt);
int radar_scan(int activate_qspn);
int radard(PACKET rpkt);
int radar_recv_reply(PACKET pkt);
void *radar_daemon(void *null);
void radar_wait_new_scan(void);

int refresh_hook_root_node(void);
